#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Nov  7 14:02:18 2018

@author: tanya
"""

import logging
import pandas as pd

from libraries.feature_engineering.in_memory_feature_engineering import StatisticalFeatures
from libraries.exception_handling import InputChecks, InputCasts
from libraries.logging.logging_utils import configure_logging

class StatisticalFeaturesOverTime(StatisticalFeatures):
    '''
    '''
    def __init__(self, data, index_cols, date_col, split_date, period_length = None, past_or_future = 'past', freq = 'days', path_to_log = None):
        '''
        '''
        configure_logging(path_to_log)
        self.logger = logging.getLogger(__name__)
        self.checks = InputChecks(logger = self.logger)
        self.casts = InputCasts(logger = self.logger)
        
        self.checks.assert_column_presence(data = data, colnames = [date_col])
        self.assert_valid_value(argname = 'past_or_future', val = past_or_future, valid_values = ['past', 'future'])
        self.assert_valid_value(argname = 'freq', val = freq, valid_values = ['seconds', 'minutes', 'hours', 'days', 'weeks', 'months', 'years'])
        
        
        if past_or_future == 'past':
            if not period_length is None:
                min_date = split_date - pd.DateOffset(**{freq : period_length})
            else:
                min_date = data[date_col].min()
            sup_date = split_date
        else:
            min_date = split_date
            if not period_length is None:
                sup_date = split_date + pd.DateOffset(**{freq : period_length})
            else: 
                sup_date = split_date + pd.DateOffset(**{freq : 1})
            
        split_date = self.casts.cast_arg_to_pandas_datetime(argname = 'split_date', val = split_date)
        data[date_col] = self.casts.cast_column_to_pandas_datetime(series = data[date_col], colname = date_col, all_or_any = 'all')    
        
            
        time_mask = (data[date_col] >= min_date) & (data[date_col] < sup_date)
        
        super(StatisticalFeaturesOverTime).__init__(data = data.loc[time_mask].reset_index(drop = True).copy(deep = True),
                                                    index_cols = index_cols,
                                                    path_to_log = path_to_log)